/*
 * Copyright (C) 2003      Martyn Russell <ginxd@btopenworld.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* 
 * $RCSfile: gj_gtk_connection_progress.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/11/09 20:42:54 $
 * 
 * $Revision: 1.38 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdlib.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_GNOME
#include <gnome.h>
#endif

#include "gj_main.h"
#include "gj_parser.h"
#include "gj_iq.h"
#include "gj_iq_requests.h"
#include "gj_roster.h"
#include "gj_translate.h"
#include "gj_me.h"
#include "gj_support.h"
#include "gj_config.h"
#include "gj_stream.h"
#include "gj_connection.h"

#include "gj_gtk_roster.h"
#include "gj_gtk_connection_progress.h"
#include "gj_gtk_connection_settings.h"
#include "gj_gtk_password_required.h"
#include "gj_gtk_support.h"


/* local structs */
enum _GjConnectionProgressState 
  {
    GjConnectionProgressStateUninitiated = 1,
    GjConnectionProgressStateIdle,
    GjConnectionProgressStateConnecting,
    GjConnectionProgressStateAuthenticationMethodRequested,
    GjConnectionProgressStateAuthenticating,
    GjConnectionProgressStateReady,
    GjConnectionProgressStateDisconnected,
    
    GjConnectionProgressStateEnd
  }; 


typedef struct 
{
  GtkWidget *window;
  
  GtkWidget *button;

  GtkWidget *label_desc1;
  GtkWidget *label_desc2;

  GtkWidget *image;

  GtkWidget *progressbar;

  gint state;

  gchar *server;
  gint16 port;
  gchar *username;
  gchar *password;
  gchar *resource;

  GjConnectionProgressCallback cb_connected;
  GjConnectionProgressCallback cb_disconnected;

  gboolean is_disconnect_planned;

  gint timeout_id;

} GjConnectionProgressWindow;


/* local functions */
static gboolean gj_gtk_cp_reload();

static gboolean gj_gtk_cp_set_state(GjConnectionProgressWindow *window, gint state);


/* 
 * states
 */

/* translations */
static const gchar *gj_gtk_cp_translate_state(gint state);

/* state machine */
static void gj_gtk_cp_state_machine(guint event_id, guint related_event_id);
static void gj_gtk_cp_state_handle_uninitiated(guint event_id, guint related_event_id);
static void gj_gtk_cp_state_handle_idle(guint event_id, guint related_event_id);
static void gj_gtk_cp_state_handle_connecting(guint event_id, guint related_event_id);
static void gj_gtk_cp_state_handle_authentication_method_requested(guint event_id, guint related_event_id);
static void gj_gtk_cp_state_handle_authenticating(guint event_id, guint related_event_id);
static void gj_gtk_cp_state_handle_ready(guint event_id, guint related_event_id);
static void gj_gtk_cp_state_handle_disconnected(guint event_id, guint related_event_id);

static void gj_gtk_cp_connection_start(GjStream stm, gint event_id);
static void gj_gtk_cp_connection_stop(GjStream stm, gint event_id);
static void gj_gtk_cp_connection_error(GjStream stm, gint event_id);

static void gj_gtk_cp_complete(gboolean success);

static gboolean gj_gtk_cp_progress_timeout_cb(GjConnectionProgressWindow *window);
static void gj_gtk_cp_enter_password_cb(gboolean success, const gchar *password);

/* gui callbacks */
static void on_button_action_clicked(GtkButton *button, GjConnectionProgressWindow *window);
static void on_realize(GtkWidget *widget, GjConnectionProgressWindow *window);
static void on_destroy(GtkWidget *widget, GjConnectionProgressWindow *window);
static gboolean on_delete_event(GtkWidget *widget, GdkEvent *event, GjConnectionProgressWindow *window);


static GjConnectionProgressWindow *current_window = NULL;


static gboolean gj_gtk_cp_reload(GjConnectionProgressWindow *window)
{
  if(!window)
    return FALSE;

  return gj_gtk_cp_load(window->cb_connected, window->cb_disconnected);
}

gboolean gj_gtk_cp_load(GjConnectionProgressCallback cb_connected,
			GjConnectionProgressCallback cb_disconnected)
{
  GladeXML *xml = NULL;
  GjConnectionProgressWindow *window = NULL;

  if(current_window != NULL)
    {
      gtk_window_present(GTK_WINDOW(current_window->window));
      return TRUE;
    }

  window = g_new0(GjConnectionProgressWindow, 1);

  window->cb_connected = cb_connected;
  window->cb_disconnected = cb_disconnected;

  window->server = g_strdup(gj_config_get_cs_server());
  window->port = gj_config_get_cs_port();
  window->username = g_strdup(gj_config_get_cs_username());
  window->password = g_strdup(gj_config_get_cs_password());
  window->resource = g_strdup(gj_config_get_cs_resource());

  window->is_disconnect_planned = FALSE;

  /* get started ! */
  gj_gtk_cp_set_state(window, GjConnectionProgressStateIdle);
  
  xml = gj_gtk_glade_get_file(gnome_jabber_get_glade_file(),
			      "connection_progress",
			      NULL,
			      "connection_progress", &window->window,
 			      "button", &window->button,
 			      "label_progress1", &window->label_desc1,
 			      "label_progress2", &window->label_desc2,
 			      "image", &window->image,
 			      "progressbar", &window->progressbar,
			      NULL);

  gj_gtk_glade_connect(xml,
		       window,
		       "connection_progress", "delete_event", on_delete_event,
		       "connection_progress", "destroy", on_destroy,
		       "connection_progress", "realize", on_realize,
		       "button", "clicked", on_button_action_clicked,
		       NULL);

  if(!xml) return FALSE;
  g_object_unref(xml);

  /* remember */
  current_window = window;

  if(gj_config_get_cs_save_password() == FALSE)
    {
      /* get password */
      gj_gtk_pr_load(FALSE, NULL, gj_gtk_cp_enter_password_cb);
    }
  else
    {
      g_message("%s: auto connecting...", __FUNCTION__);
      gtk_widget_show_all(window->window);
      gj_gtk_cp_state_machine(0,0);
    }

  return TRUE;
} 

static void gj_gtk_cp_enter_password_cb(gboolean success, const gchar *password)
{
  /* this could be because:
     - user closed the dialog
     - user clicked ok, and the password wrong,
     - user clicked cancel     */

  if(!current_window)
    return;

  if(success == FALSE)
    {
      g_message("%s: password dialog cancelled, not progressing further.", __FUNCTION__);
      gtk_widget_destroy(current_window->window);
      return;
    }

  /* set password */
  if(current_window->password != NULL)
    {
      g_free(current_window->password);
      current_window->password = NULL;
    }

  current_window->password = g_strdup(password);

  /* start the state machine */
  gtk_widget_show_all(current_window->window);
  gj_gtk_cp_state_machine(0, 0);
}

/* sets */
static gboolean gj_gtk_cp_set_state(GjConnectionProgressWindow *window, gint state)
{
  if(!window)
    return FALSE;

  if(state < 1 || state >= GjConnectionProgressStateEnd)
    {
      g_warning("%s: state:%d was out of range", __FUNCTION__, state);
      return FALSE;
    }

  g_message("+++++++++++++++++++++++++++++++++++++++++");
  g_message("++++ set state:%d->'%s'", state, gj_gtk_cp_translate_state(state));
  g_message("+++++++++++++++++++++++++++++++++++++++++");

  window->state = state;
  return TRUE;
}

static const gchar *gj_gtk_cp_translate_state(gint state)
{
  switch(state)
    {
    case GjConnectionProgressStateUninitiated: return _("uninitiated");
    case GjConnectionProgressStateIdle: return _("idle");
    case GjConnectionProgressStateConnecting: return _("connecting");
    case GjConnectionProgressStateAuthenticationMethodRequested: return _("authentication method requested");
    case GjConnectionProgressStateAuthenticating: return _("authenticating");
    case GjConnectionProgressStateReady: return _("ready");
    case GjConnectionProgressStateDisconnected: return _("disconnected");
    }  

  return "unknown";
}

static void gj_gtk_cp_state_machine(guint event_id, guint related_event_id)
{
  if(!current_window)
    return;

  g_message("*****************************************");
  g_message("**** entered state:%d->'%s'", 
	    current_window->state, 
	    gj_gtk_cp_translate_state(current_window->state));
  g_message("*****************************************");

  switch(current_window->state)
    {
    case GjConnectionProgressStateUninitiated:
      gj_gtk_cp_state_handle_uninitiated(event_id, related_event_id); break;
    case GjConnectionProgressStateIdle:
      gj_gtk_cp_state_handle_idle(event_id, related_event_id); break;
    case GjConnectionProgressStateConnecting:
      gj_gtk_cp_state_handle_connecting(event_id, related_event_id); break;
    case GjConnectionProgressStateAuthenticationMethodRequested:
      gj_gtk_cp_state_handle_authentication_method_requested(event_id, related_event_id); break;
    case GjConnectionProgressStateAuthenticating:
      gj_gtk_cp_state_handle_authenticating(event_id, related_event_id); break;
    case GjConnectionProgressStateReady:
      gj_gtk_cp_state_handle_ready(event_id, related_event_id); break;
    case GjConnectionProgressStateDisconnected:
      gj_gtk_cp_state_handle_disconnected(event_id, related_event_id); break;
    }  
}

static void gj_gtk_cp_state_handle_uninitiated(guint event_id, guint related_event_id)
{
  if(!current_window)
    return;
}

static void gj_gtk_cp_state_handle_idle(guint event_id, guint related_event_id)
{
  GjStream stm = NULL;

  if(!current_window)
    return;

  gj_me_set_is_connected(FALSE);
  gj_me_set_uptime_stop();

  /* set up progress bar */
  gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_INFO,GTK_ICON_SIZE_DIALOG);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(current_window->progressbar),(gdouble)1/4);
  gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Connecting..."));
  gj_gtk_set_label_with_args(current_window->label_desc2, _("server '%s' on port %d"),
			     current_window->server, 
			     current_window->port);

  /* create stream */
  if((stm = gj_stream_new(current_window->server)) == NULL)
    {
      gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Error"));
      gj_gtk_set_label_with_args(current_window->label_desc2, _("Could not create new stream."));

      g_warning("%s: could not create new stream", __FUNCTION__);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }
 
  /* set struct info */
  gj_stream_set_port(stm, current_window->port);
  gj_stream_set_connected_cb(stm, gj_gtk_cp_connection_start);
  gj_stream_set_disconnected_cb(stm, gj_gtk_cp_connection_stop);
  gj_stream_set_error_cb(stm, gj_gtk_cp_connection_error);

  if(gj_stream_build(stm) == FALSE)
    {
      gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Error"));
      gj_gtk_set_label_with_args(current_window->label_desc2, _("Could not build stream."));

      g_warning("%s: could not build new stream.", __FUNCTION__);

      /* finish */
      gj_gtk_cp_complete(FALSE);
      return;
    }

  if(gj_stream_open(stm) == FALSE)
    {
      gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s..</b>", _("Connection Error"));
      gj_gtk_set_label_with_args(current_window->label_desc2, _("Unable to open stream."));

      /* finish */
      gj_gtk_cp_complete(FALSE);
      return;
    }

  /* add timeout for pulsing */
  current_window->timeout_id = gtk_timeout_add(25, (GSourceFunc)gj_gtk_cp_progress_timeout_cb, current_window);

  /* set state */
  gj_gtk_cp_set_state(current_window, GjConnectionProgressStateConnecting);
}

static void gj_gtk_cp_state_handle_connecting(guint event_id, guint related_event_id)
{
  gchar *detail = NULL;
  gchar *jid = NULL;

  if(!current_window)
    return;

  /* create jid */
  jid = g_strdup_printf("%s@%s%s%s", 
			current_window->username, 
			current_window->server, 
			current_window->resource?"/":"",
			current_window->resource?current_window->resource:"");

  /* set the fact that we are now connected */
  gj_me_set_is_connected(TRUE);
  gj_me_set_uptime_start();

  gj_me_set_connected_host(current_window->server);
  gj_me_set_connected_port(current_window->port);
  gj_me_set_connected_jid(jid);

  gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_INFO,GTK_ICON_SIZE_DIALOG);
  gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s...</b>", _("Requesting Authentication Method"));
  gj_gtk_set_label_with_args(current_window->label_desc2, _("username '%s'"), current_window->username);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(current_window->progressbar),(gdouble)2/4);
  g_message("%s: response received, event_id:%d", __FUNCTION__, event_id);

  /* clean up */
  if(detail != NULL) g_free(detail);
  if(jid != NULL) g_free(jid);

  /* NOTE: no need to free the iq by event id because there was no iq
     created to generate this response! */

  /* next request */
  gj_iq_request_auth_method(current_window->username, gj_gtk_cp_state_machine, NULL);

  /* set state */
  gj_gtk_cp_set_state(current_window, GjConnectionProgressStateAuthenticationMethodRequested);
}

static void gj_gtk_cp_state_handle_authentication_method_requested(guint event_id, guint related_event_id)
{
  gchar *detail = NULL;
  
  const gchar *password_plain = NULL;
  gchar *password_digest = NULL;

  const gchar *session_id = NULL;

  GjStream stm = NULL;

  if(event_id < 0)
    {
      gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Request for Authentication Method Failed."));
      gj_gtk_set_label_with_args(current_window->label_desc2, _("Invalid Response."));
      g_warning("%s: failed to continue with response to authorisation, event_id:%d was < 0", __FUNCTION__, event_id);

      /* set state */
      gj_gtk_cp_set_state(current_window, GjConnectionProgressStateIdle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }

  detail = g_strdup_printf(_("username '%s'"), current_window->username);

  gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_INFO,GTK_ICON_SIZE_DIALOG);
  gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Requesting Authentication..."));
  gj_gtk_set_label_with_args(current_window->label_desc2, detail);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(current_window->progressbar),(gdouble)3/4);
  g_message("%s: received available authentication methods, event_id:%d", __FUNCTION__, event_id);

  /* clean up */
  g_free(detail);

  /* find stream */
  if((stm = gj_stream_find_by_to(current_window->server)) == NULL)
    {
      gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Request for Authentication Method Failed."));
      gj_gtk_set_label_with_args(current_window->label_desc2, _("Internal Error."));
      g_warning("%s: failed to continue with response to authorisation, could not find stream for connection to '%s'", __FUNCTION__, current_window->server);

      /* clean up */
      gj_iq_free_all_related_by_event_id(event_id);
  
      /* set state */
      gj_gtk_cp_set_state(current_window, GjConnectionProgressStateIdle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }

  /* get session id */
  session_id = gj_stream_get_id(stm);

  /* clean up */
  gj_iq_free_all_related_by_event_id(event_id);
 
  /* next request */
  if(session_id == NULL || gj_config_get_cs_use_plain_text_password() == TRUE)
    {
      /* use plain text */
      gj_iq_request_auth_with_plain_password(current_window->username,
					     current_window->password,
					     current_window->resource,
					     gj_gtk_cp_state_machine,
					     NULL);
    }
  else
    {
      /* use digest (SHA1) */
      password_plain = current_window->password;
      password_digest = gj_get_sha_by_session_id_and_password(session_id, password_plain);

      gj_iq_request_auth_with_digest_password(current_window->username,
					      password_digest,
					      current_window->resource,
					      gj_gtk_cp_state_machine,
					      NULL);
      
      if(password_digest != NULL) g_free(password_digest);
    }

  /* set state */
  gj_gtk_cp_set_state(current_window, GjConnectionProgressStateAuthenticating);
}

static void gj_gtk_cp_state_handle_authenticating(guint event_id, guint related_event_id)
{
  GjInfoquery iq = NULL;
  GjInfoquery iq_related = NULL;

  guint event_idRelated = 0;

  xmlChar *type = NULL;
  xmlNodePtr node = NULL;

  if(!current_window)
    return;

  if((iq = gj_iq_find(event_id)) == NULL)
    {
      gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(current_window->progressbar),(gdouble)4/4);
      gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Failed to authenticate."));
      gj_gtk_set_label_with_args(current_window->label_desc2, _("Invalid response."));
    
      g_warning("%s: failed to find IQ based on event_id:%d", __FUNCTION__, event_id);

      /* clean up */
      gj_iq_free_all_related_by_event_id(event_id);

      /* set state */
      gj_gtk_cp_set_state(current_window, GjConnectionProgressStateIdle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }

  event_idRelated = gj_iq_get_related_id(iq);
  if((iq_related = gj_iq_find(event_idRelated)) == NULL)
    {      
      gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(current_window->progressbar),(gdouble)4/4);
      gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Failed to authenticate."));
      gj_gtk_set_label_with_args(current_window->label_desc2, _("Invalid data in related request."));

      g_warning("%s: failed to find IQ based on event_idRelated:%d", __FUNCTION__, event_idRelated);

      /* clean up */
      gj_iq_free_all_related_by_event_id(event_id);

      /* set state */
      gj_gtk_cp_set_state(current_window, GjConnectionProgressStateIdle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }

  node = gj_iq_get_xml(iq_related);
  type = gj_parser_find_by_attr_and_ret(node, (guchar*)"type");

  if(type != NULL && xmlStrcmp(type, (guchar*)"error") == 0)
    {
      gchar *detail = NULL;

      xmlChar *errorCode = NULL;
      xmlChar *errorMessage = NULL;

      if((node = gj_parser_find_by_node(node, (guchar*)"error")) == NULL)
	{
	  g_warning("%s: failed to find node 'error'", __FUNCTION__);

	  /* clean up */
	  gj_iq_free_all_related_by_event_id(event_id);

	  /* set state */
	  gj_gtk_cp_set_state(current_window, GjConnectionProgressStateIdle);

	  /* finish */
	  gj_gtk_cp_complete(FALSE);

	  return;
	}

      if((errorCode = gj_parser_find_by_attr_and_ret(node, (guchar*)"code")) != NULL)
	detail = g_strdup_printf("%s:%s", _("Error"), errorCode);

      if((errorMessage = gj_parser_find_by_node_and_ret(node, (guchar*)"error")) != NULL) 
	if(detail != NULL) detail = g_strdup_printf("%s - %s",detail,errorMessage);

      gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(current_window->progressbar),(gdouble)4/4);
      gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Failed to authenticate."));
      gj_gtk_set_label_with_args(current_window->label_desc2, detail);
      if(detail != NULL) g_free(detail);

      g_warning("%s: IQ type is error, authentication failed for event_id:%d", __FUNCTION__, event_id);

      /* clean up */
      gj_iq_free_all_related_by_event_id(event_id);

      /* set state */
      gj_gtk_cp_set_state(current_window, GjConnectionProgressStateIdle);

      /* finish */
      gj_gtk_cp_complete(FALSE);

      return;
    }
  
  gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_INFO,GTK_ICON_SIZE_DIALOG);
  gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(current_window->progressbar),(gdouble)4/4);
  gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Received Authentication."));
  gj_gtk_set_label_with_args(current_window->label_desc2, _("Cleaning up..."));

  g_message("%s: received authentication response, event_id:%d", __FUNCTION__, event_id);

  /* clean up */
  gj_iq_free_all_related_by_event_id(event_id);

  /* set state */
  gj_gtk_cp_set_state(current_window, GjConnectionProgressStateReady);

  /* finish */
  gj_gtk_cp_complete(TRUE);
}

static void gj_gtk_cp_state_handle_ready(guint event_id, guint related_event_id)
{
}

static void gj_gtk_cp_state_handle_disconnected(guint event_id, guint related_event_id)
{
  GjConnectionProgressCallback callback = NULL;

  if(!current_window)
    return;

  /* set state */
  gj_me_set_is_connected(FALSE);
  gj_me_set_uptime_stop();

  /* call callback */
  if((callback = current_window->cb_disconnected) != NULL)
    (callback)();

  /* should we reconnect ? */
  if(current_window->is_disconnect_planned)
    {
      gtk_widget_destroy(current_window->window);
      return;
    }
     
  if(gj_config_get_cs_auto_reconnect_on_disconnect() == TRUE && gj_config_get_cs_save_password() == TRUE)
    gj_gtk_cp_reload(current_window);
  else
    gtk_widget_destroy(current_window->window);
}

static void gj_gtk_cp_connection_start(GjStream stm, gint event_id)
{
  gj_gtk_cp_state_machine(event_id, 0);
}

static void gj_gtk_cp_connection_stop(GjStream stm, gint event_id)
{
  if(!current_window)
    return;

  if(gj_stream_get_is_disconnect_planned(stm) == TRUE)
    current_window->is_disconnect_planned = TRUE;

  gj_gtk_cp_set_state(current_window, GjConnectionProgressStateDisconnected);
  gj_gtk_cp_state_machine(event_id, 0);
}

static void gj_gtk_cp_connection_error(GjStream stm, gint event_id)
{
  if(!current_window)
    return;

  /* display error */
  gtk_image_set_from_stock(GTK_IMAGE(current_window->image),GTK_STOCK_DIALOG_ERROR,GTK_ICON_SIZE_DIALOG);
  gj_gtk_set_label_with_args(current_window->label_desc1, "<b>%s</b>", _("Failed to connect."));
  gj_gtk_set_label_with_args(current_window->label_desc2, (gchar*)gj_connection_connect_translate_status(event_id));
 
  gj_gtk_cp_complete(FALSE);
}

static void gj_gtk_cp_complete(gboolean success)
{
  GjConnectionProgressCallback callback = NULL;

  if(!current_window)
    return;

  /* call callback */
  if(success == TRUE)
    {
      if((callback = current_window->cb_connected) != NULL)
	(callback)();

      /* close the window */
      gtk_widget_hide(current_window->window); 
    }
  else
    {
      gj_me_set_is_connected(FALSE);
      gj_me_set_uptime_stop();
      
      g_source_remove(current_window->timeout_id);
      gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(current_window->progressbar),(gdouble)1/1);
      
      if((callback = current_window->cb_disconnected) != NULL)
	(callback)();
    }
}

static gboolean gj_gtk_cp_progress_timeout_cb(GjConnectionProgressWindow *window)
{
  if(!window)
    return FALSE;

  if(window->state == GjConnectionProgressStateAuthenticating)
    return FALSE;

  gtk_progress_bar_pulse(GTK_PROGRESS_BAR(window->progressbar));
  
  /* As this is a timeout function, return TRUE so that it
   * continues to get called */
  return TRUE;
} 

/*
 * GUI events
 */
static void on_button_action_clicked(GtkButton *button, GjConnectionProgressWindow *window)
{
  if(!window)
    return;

#if 0
  if(window->server != NULL && window->port > 0)
    {
      GjStream stm = NULL;

      if((stm = streamFindByTo(window->server)) != NULL)
	streamFree(stm);
    }
#endif

  gtk_widget_destroy(window->window);
}

static void on_realize(GtkWidget *widget, GjConnectionProgressWindow *window)
{
  if(!window)
    return;

  /* if we do it on load, the window is hidden, and we get warnings */
  gdk_window_set_decorations (GDK_WINDOW(window->window->window), 
			      GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU);
}

static void on_destroy(GtkWidget *widget, GjConnectionProgressWindow *window)
{
  current_window = NULL;
  if(!window)
    return;

  if(window->server != NULL)
    g_free(window->server);
  if(window->username != NULL)
    g_free(window->username);
  if(window->password != NULL)
    g_free(window->password);
  if(window->resource != NULL)
    g_free(window->resource);

  g_free(window);
}

static gboolean on_delete_event(GtkWidget *widget, GdkEvent *event,  GjConnectionProgressWindow *window)
{
  /* return TRUE so it doesnt close */
  return TRUE;
}



#ifdef __cplusplus
}
#endif
 
