/*
 * Copyright (C) 2003      Martyn Russell <ginxd@btopenworld.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* 
 * $RCSfile: gj_iq_requests.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/11/08 00:05:48 $
 * 
 * $Revision: 1.19 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>

#include "gj_iq.h"
#include "gj_iq_requests.h"
#include "gj_iq_responses.h"
#include "gj_roster.h"
#include "gj_parser.h"
#include "gj_stream.h"


/*
 * ACTUAL REQUESTS from outside this module, eg:
 *
 *  - Greeting 
 *  - Auth Method
 *  - Auth
 *  - Roster
 *  - RosterRemove
 *  - RosterAdd
 *  - PasswordChange
 *  - UserChange (name, group...) 
 *  - VCardChange
 *  - VCard
 *  - Version
 *  - Time
 *  - Last
 *  - Agents
 *  - Search
 *  ...
 *
 */

gboolean gj_iq_request_goodbye(const gchar *to)
{
  gj_stream_close_by_to(to);
  
  return TRUE;
}

gboolean gj_iq_request_auth_method(const gchar *username, 
				   GjInfoqueryCallback cb,
				   void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  /* checks */
  if(username == NULL)
    {
      g_warning("%s: username was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)  
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(username, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:auth", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"username", (guchar*)s1);

  /* set struct info */
  gj_iq_set_callback(iq,cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq, node) == FALSE)
    {
      g_warning("%s: failed build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_auth_with_plain_password(const gchar *username, 
						const gchar *password, 
						const gchar *resource, 
						GjInfoqueryCallback cb,
						void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;
  gchar *s2 = NULL;
  gchar *s3 = NULL;

  /* checks */
  if(username == NULL)
    {
      g_warning("%s: username was NULL", __FUNCTION__);
      return FALSE;
    }

  if(password == NULL)
    {
      g_warning("%s: password was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(username, -1);
  s2 = g_markup_escape_text(password, -1);
  s3 = g_markup_escape_text(resource, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:auth", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"username", (guchar*)s1);
  sub = xmlNewChild(node, ns, (guchar*)"password", (guchar*)s2);
  sub = xmlNewChild(node, ns, (guchar*)"resource", (guchar*)s3);

  /* set struct info */
  gj_iq_set_callback(iq,cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);
  if(s2 != NULL) g_free(s2);
  if(s3 != NULL) g_free(s3);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_auth_with_digest_password(const gchar *username, 
						 const gchar *digest, 
						 const gchar *resource, 
						 GjInfoqueryCallback cb,
						 void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;
  gchar *s2 = NULL;
  gchar *s3 = NULL;

  /* checks */   
  if(username == NULL)
    {
      g_warning("%s: username was NULL", __FUNCTION__);
      return FALSE;
    }

  if(digest == NULL)
    {
      g_warning("%s: digest was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: information query was NULL", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(username, -1);
  s2 = g_markup_escape_text(digest, -1);
  s3 = g_markup_escape_text(resource, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:auth", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"username", (guchar*)s1);
  sub = xmlNewChild(node, ns, (guchar*)"digest", (guchar*)s2);
  sub = xmlNewChild(node, ns, (guchar*)"resource", (guchar*)s3);

  /* set struct info */
  gj_iq_set_callback(iq,cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);
  if(s2 != NULL) g_free(s2);
  if(s3 != NULL) g_free(s3);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_roster(GjInfoqueryCallback cb, void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNsPtr ns = NULL;

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: query was NULL", __FUNCTION__);
      return FALSE;
    }

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:roster", NULL);
 
  /* set struct info */
  gj_iq_set_callback(iq, cb);
  
  if(userdata != NULL)
    gj_iq_set_userdata(iq, userdata);
    
  if(gj_iq_build(iq, node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_roster_remove_contact(const gchar *jid, GjInfoqueryCallback cb, void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL; 
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(jid, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:roster", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"item", NULL);
  xmlSetProp(sub, (guchar*)"jid", (guchar*)s1);
  xmlSetProp(sub, (guchar*)"subscription", (guchar*)"remove");
  
  /* set struct info */
  gj_iq_set_callback(iq, cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq, node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);  
}

gboolean gj_iq_request_roster_add_contact(const gchar *jid, 
					  const gchar *name, 
					  GList *groups, 
					  GjInfoqueryCallback cb,
					  void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL; 
  xmlNsPtr ns = NULL;

  gint i = 0;

  gchar *s1 = NULL;
  gchar *s2 = NULL;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(name == NULL)
    {
      g_warning("%s: name was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(jid, -1);
  s2 = g_markup_escape_text(name, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:roster", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"item", NULL);
  xmlSetProp(sub, (guchar*)"jid", (guchar*)s1);
  xmlSetProp(sub, (guchar*)"name", (guchar*)s2);

  if(groups != NULL)
    {
      for(i=0;i<g_list_length(groups);i++)
	{
	  gchar *group = NULL;
	  gchar *s4 = NULL;

	  if((group = (gchar*) g_list_nth_data(groups, i)) == NULL)
	    continue;

	  s4 = g_markup_escape_text(group, -1);
	  xmlNewChild(sub, ns, (guchar*)"group", (guchar*)s4);

	  if(s4 != NULL) g_free(s4);
	}
    }

  /* set struct info */
  gj_iq_set_callback(iq,cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);
  if(s2 != NULL) g_free(s2);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);  
}

gboolean gj_iq_request_password_change(const gchar *to, 
				       const gchar *password, 
				       GjInfoqueryCallback cb,
				       void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;
  gchar *s2 = NULL;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(password == NULL)
    {
      g_warning("%s: password was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);
  s2 = g_markup_escape_text(password, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:register", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"password", (guchar*)s2);
 
  /* set struct info */
  gj_iq_set_to(iq, s1);
  gj_iq_set_callback(iq, cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);
  if(s2 != NULL) g_free(s2);

  if(gj_iq_build(iq, node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_user_change(const gchar *jid, 
				   const gchar *name, 
				   GList *groups, 
				   GjInfoqueryCallback cb,
				   void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;
  gchar *s2 = NULL;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(name == NULL)
    {
      g_warning("%s: name was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(jid, -1);
  s2 = g_markup_escape_text(name, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:roster", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"item", NULL);
  xmlSetProp(sub, (guchar*)"jid", (guchar*)jid);
  xmlSetProp(sub, (guchar*)"name", (guchar*)name);

  /* go through groups list and add each item */
  if(groups != NULL)
    {
      gint i = 0;
      
      for(i=0;i<g_list_length(groups);i++)
	{
	  gchar *g = (gchar*) g_list_nth_data(groups, i);
	  gchar *s3 = NULL;

	  if(g == NULL) continue;

	  s3 = g_markup_escape_text(g, -1);
	  xmlNewChild(sub, ns, (guchar*)"group", (guchar*)s3);

	  if(s3 != NULL) g_free(s3);
	}
    }
 
  /* set struct info */
  gj_iq_set_callback(iq,cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);
  if(s2 != NULL) g_free(s2);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_vcard_change(const gchar *fn,
				    const gchar *given,
				    const gchar *family,
				    const gchar *nickname,
				    const gchar *url,
				    const gchar *tel,
				    const gchar *email,
				    const gchar *desc,
				    const gchar *bday,
				    const gchar *street,
				    const gchar *extadd,
				    const gchar *locality,
				    const gchar *region,
				    const gchar *pcode,
				    const gchar *country,
				    const gchar *orgname,
				    const gchar *orgunit,
				    const gchar *title,
				    const gchar *role,
				    GjInfoqueryCallback cb,
				    void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub1 = NULL;
  xmlNodePtr sub2 = NULL;
  xmlNsPtr ns = NULL;

  gchar *s01 = NULL;
  gchar *s02 = NULL;
  gchar *s03 = NULL;
  gchar *s04 = NULL;
  gchar *s05 = NULL;
  gchar *s06 = NULL;
  gchar *s07 = NULL;
  gchar *s08 = NULL;
  gchar *s09 = NULL;
  gchar *s10 = NULL;
  gchar *s11 = NULL;
  gchar *s12 = NULL;
  gchar *s13 = NULL;
  gchar *s14 = NULL;
  gchar *s15 = NULL;
  gchar *s16 = NULL;
  gchar *s17 = NULL;
  gchar *s18 = NULL;
  gchar *s19 = NULL;

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s01 = g_markup_escape_text(fn, -1);
  s02 = g_markup_escape_text(family, -1);
  s03 = g_markup_escape_text(given, -1);
  s04 = g_markup_escape_text(nickname, -1);
  s05 = g_markup_escape_text(url, -1);
  s06 = g_markup_escape_text(tel, -1);
  s07 = g_markup_escape_text(email, -1);
  s08 = g_markup_escape_text(desc, -1);
  s09 = g_markup_escape_text(bday, -1);
  s10 = g_markup_escape_text(street, -1);
  s11 = g_markup_escape_text(extadd, -1);
  s12 = g_markup_escape_text(locality, -1);
  s13 = g_markup_escape_text(region, -1);
  s14 = g_markup_escape_text(pcode, -1);
  s15 = g_markup_escape_text(country, -1);
  s16 = g_markup_escape_text(orgname, -1);
  s17 = g_markup_escape_text(orgunit, -1);
  s18 = g_markup_escape_text(title, -1);
  s19 = g_markup_escape_text(role, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"vcard");
  ns = xmlNewNs(node, (guchar*)"vcard-temp", NULL);
  sub1 = xmlNewChild(node, ns, (guchar*)"fn", (guchar*)s01);

  sub1 = xmlNewChild(node, ns, (guchar*)"n", NULL);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"family", (guchar*)s02);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"given", (guchar*)s03);

  sub1 = xmlNewChild(node, ns, (guchar*)"nickname", (guchar*)s04);
  sub1 = xmlNewChild(node, ns, (guchar*)"url", (guchar*)s05);
  sub1 = xmlNewChild(node, ns, (guchar*)"tel", (guchar*)s06);
  sub1 = xmlNewChild(node, ns, (guchar*)"email", (guchar*)s07);
  sub1 = xmlNewChild(node, ns, (guchar*)"desc", (guchar*)s08);
  sub1 = xmlNewChild(node, ns, (guchar*)"bday", (guchar*)s09);

  sub1 = xmlNewChild(node, ns, (guchar*)"adr", NULL);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"street", (guchar*)s10);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"extadd", (guchar*)s11);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"locality", (guchar*)s12);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"region", (guchar*)s13);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"pcode", (guchar*)s14);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"country", (guchar*)s15);

  sub1 = xmlNewChild(node, ns, (guchar*)"org", NULL);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"orgname", (guchar*)s16);
  sub2 = xmlNewChild(sub1, ns, (guchar*)"orgunit", (guchar*)s17);

  sub1 = xmlNewChild(node, ns, (guchar*)"title", (guchar*)s18);
  sub1 = xmlNewChild(node, ns, (guchar*)"role", (guchar*)s19);

  /* set struct info */
  gj_iq_set_callback(iq,cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s01 != NULL) g_free(s01);
  if(s02 != NULL) g_free(s02);
  if(s03 != NULL) g_free(s03);
  if(s04 != NULL) g_free(s04);
  if(s05 != NULL) g_free(s05);
  if(s06 != NULL) g_free(s06);
  if(s07 != NULL) g_free(s07);
  if(s08 != NULL) g_free(s08);
  if(s09 != NULL) g_free(s09);
  if(s10 != NULL) g_free(s10);
  if(s11 != NULL) g_free(s11);
  if(s12 != NULL) g_free(s12);
  if(s13 != NULL) g_free(s13);
  if(s14 != NULL) g_free(s14);
  if(s15 != NULL) g_free(s15);
  if(s16 != NULL) g_free(s16);
  if(s17 != NULL) g_free(s17);
  if(s18 != NULL) g_free(s18);
  if(s19 != NULL) g_free(s19);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_vcard(const gchar *to, 
			     GjInfoqueryCallback cb, 
			     void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"vcard");
  ns = xmlNewNs(node, (guchar*)"vcard-temp", NULL);
 
  /* set struct info */
  gj_iq_set_to(iq,s1);
  gj_iq_set_callback(iq,cb);
  gj_iq_set_userdata(iq,userdata);

  /* set struct info */
  gj_iq_set_callback(iq,cb);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_version(const gchar *to, 
			       GjInfoqueryCallback cb, 
			       void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:version", NULL);
 
  /* set struct info */
  gj_iq_set_to(iq,to);
  gj_iq_set_callback(iq,cb);
  gj_iq_set_userdata(iq,userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq, node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_time(const gchar *to, 
			    GjInfoqueryCallback cb, 
			    void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:time", NULL);
 
  /* set struct info */
  gj_iq_set_to(iq,s1);
  gj_iq_set_callback(iq,cb);
  gj_iq_set_userdata(iq,userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_last(const gchar *to, 
			    GjInfoqueryCallback cb, 
			    void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:last", NULL);
 
  /* set struct info */
  gj_iq_set_to(iq,s1);
  gj_iq_set_callback(iq,cb);
  gj_iq_set_userdata(iq,userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_register_user(const gchar *to, 
				     const gchar *username, 
				     const gchar *password, 
				     GjInfoqueryCallback cb,
				     void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;
  gchar *s2 = NULL;
  gchar *s3 = NULL;

  /* checks */
  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(username == NULL)
    {
      g_warning("%s: username was NULL", __FUNCTION__);
      return FALSE;
    }

  if(password == NULL)
    {
      g_warning("%s: password was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);
  s2 = g_markup_escape_text(username, -1);
  s3 = g_markup_escape_text(password, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:register", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"username", (guchar*)s2);
  sub = xmlNewChild(node, ns, (guchar*)"password", (guchar*)s3);

  /* set struct info */
  gj_iq_set_callback(iq, cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);
  gj_iq_set_to(iq, s1);

  /* clean up */
  if(s1 != NULL) g_free(s1);
  if(s2 != NULL) g_free(s2);
  if(s3 != NULL) g_free(s3);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_register(const gchar *to, 
				GjInfoqueryCallback cb,
				void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:register", NULL);
 
  /* set struct info */
  gj_iq_set_to(iq, s1);
  gj_iq_set_callback(iq, cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_registration_requirements(const gchar *to, 
						 GjInfoqueryCallback cb,
						 void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:register", NULL);
 
  /* set struct info */
  gj_iq_set_to(iq, to);
  gj_iq_set_callback(iq, cb);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_registration(const gchar *to, 
				    GjInfoqueryCallback cb, 
				    void *userdata,
				    ...)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  va_list args;
  guchar *key = NULL;
  guchar *value = NULL;

  gchar *s1 = NULL;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:register", NULL);

  va_start(args, userdata);
  key = (guchar*)va_arg(args, char*);

  for(; ; key = (guchar*)va_arg(args, char*))
    {
      gchar *s2 = NULL;
      gchar *s3 = NULL;

      if(key == NULL) break;
      
      value = (guchar*)va_arg(args, char*);
      if(g_utf8_strlen((gchar*)value,-1) < 1) continue;

      s2 = g_markup_escape_text((gchar*)key, -1);
      s3 = g_markup_escape_text((gchar*)value, -1);
      
      sub = xmlNewChild(node, ns, (guchar*)s2, (guchar*)s3);

      if(s2 != NULL) g_free(s2);
      if(s3 != NULL) g_free(s3);
    }

  va_end(args);
 
  /* set struct info */
  gj_iq_set_to(iq,s1);
  gj_iq_set_callback(iq,cb);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_unregister(const gchar *to, 
				  GjInfoqueryCallback cb,
				  void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  /* checks */
  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeSet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(to, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query"); 
  ns = xmlNewNs(node, (guchar*)"jabber:iq:register", NULL);
  sub = xmlNewChild(node, ns, (guchar*)"remove", NULL);

  /* set struct info */
  gj_iq_set_callback(iq, cb);
  gj_iq_set_to(iq, s1);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_agents(const gchar *server, 
			      GjInfoqueryCallback cb,
			      void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  /* checks */
  if(server == NULL)
    {
      g_warning("%s: server was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(server, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:agents", NULL);
 
  /* set struct info */
  gj_iq_set_callback(iq,cb);
  gj_iq_set_to(iq,s1);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_search_requirements(const gchar *server, 
					   GjInfoqueryCallback cb,
					   void *userdata)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNsPtr ns = NULL;

  gchar *s1 = NULL;

  /* checks */
  if(server == NULL)
    {
      g_warning("%s: server was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(server, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:search", NULL);
 
  /* set struct info */
  gj_iq_set_callback(iq,cb);
  gj_iq_set_to(iq,s1);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

gboolean gj_iq_request_search(const gchar *server, 
			      GjInfoqueryCallback cb, 
			      void *userdata, 
			      ...)
{
  GjInfoquery iq = NULL;

  xmlNodePtr node = NULL;
  xmlNodePtr sub = NULL;
  xmlNsPtr ns = NULL;

  va_list args;
  guchar *key = NULL;
  guchar *value = NULL;

  gchar *s1 = NULL;

  /* checks */
  if(server == NULL)
    {
      g_warning("%s: server was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: cb was NULL", __FUNCTION__);
      return FALSE;
    }

  if((iq = gj_iq_new(GjInfoqueryTypeGet)) == NULL)
    {
      g_warning("%s: infoquery could not be created", __FUNCTION__);
      return FALSE;
    }

  s1 = g_markup_escape_text(server, -1);

  /* set xml */
  node = xmlNewNode(NULL, (guchar*)"query");
  ns = xmlNewNs(node, (guchar*)"jabber:iq:search", NULL);

  va_start(args, userdata);
  key = (guchar*)va_arg(args, char*);

  for(; ; key = (guchar*)va_arg(args, char*))
    {
      gchar *s2 = NULL;
      gchar *s3 = NULL;

      if(key == NULL) break;
      value = (guchar*)va_arg(args, char*);
      
      if(g_utf8_strlen((gchar*)value, -1) < 1) continue;

      s2 = g_markup_escape_text((gchar*)key, -1);
      s3 = g_markup_escape_text((gchar*)value, -1);
  
      sub = xmlNewChild(node, ns, (guchar*)s2, (guchar*)s3);

      if(s2 != NULL) g_free(s2);
      if(s3 != NULL) g_free(s3);
    }
  
  va_end(args);
 
  /* set struct info */
  gj_iq_set_callback(iq,cb);
  gj_iq_set_to(iq,s1);
  if(userdata != NULL) gj_iq_set_userdata(iq, userdata);

  /* clean up */
  if(s1 != NULL) g_free(s1);

  if(gj_iq_build(iq,node) == FALSE)
    {
      g_warning("%s: failed to build query.", __FUNCTION__);
      return FALSE;
    }

  return gj_iq_send(iq);
}

#ifdef __cplusplus
}
#endif
