/*
 * Copyright (C) 2003      Martyn Russell <ginxd@btopenworld.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* 
 * $RCSfile: gj_registration.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/10/30 20:38:28 $
 * 
 * $Revision: 1.17 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include "gj_registration.h"
#include "gj_parser.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_iq.h"
#include "gj_iq_requests.h"

static void gj_registration_init();

/* registration list */
static gboolean gj_registration_list_add(GjRegistration rgr);
static gboolean gj_registration_list_del(GjRegistration rgr);

/* registration item */
static void gj_registration_destroy(GjRegistration rgr);

/* parsing functions */
static void gj_registration_cb(guint event_id, guint related_event_id);
static GjRegistration gj_registration_parse(xmlNodePtr parent);


struct t_registration
{
  gchar *id;
  gchar *to;
  gchar *from;

  gchar *instructions;

  gchar *username;
  gchar *password;
  gchar *email;

  gboolean require_username;
  gboolean require_password;
  gboolean require_email;
  gboolean require_nickname;

  gchar *key;

  GjRegistrationCallback cb; 
};


static GList *registration_list = NULL;
static GjRegistrationRequestCallback registration_cb = NULL;

static void gj_registration_init()
{
  static gboolean initiated = FALSE;
  
  if(initiated == TRUE)
    return;

  initiated = TRUE;
}

static gboolean gj_registration_list_add(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  registration_list = g_list_append(registration_list, rgr);
  return TRUE;
}

static gboolean gj_registration_list_del(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  registration_list = g_list_remove(registration_list, rgr);
  return TRUE;
}

GList *gj_registration_list_get()
{
  return registration_list;
}

/* 
 * item functions
 */
GjRegistration gj_registration_new()
{
  GjRegistration rgr = NULL;

  gj_registration_init();

  if((rgr = g_new(struct t_registration,1)))
    {
      rgr->id = NULL;
      rgr->to = NULL;
      rgr->from = NULL;
      
      rgr->instructions = NULL;

      rgr->username = NULL;
      rgr->password = NULL;
      rgr->email = NULL;

      rgr->require_username = FALSE;
      rgr->require_password = FALSE;
      rgr->require_email = FALSE;
      rgr->require_nickname = FALSE;

      rgr->key = NULL;
      
      rgr->cb = NULL;

      gj_registration_list_add(rgr);
    }
  
  return rgr;  
}

void gj_registration_free(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return;
    }

  /* free mem */
  gj_registration_list_del(rgr);  

  /* destroy members */
  gj_registration_destroy(rgr);

  /* free structure */
  g_free(rgr);
}

static void gj_registration_destroy(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return;
    }
  
  if(rgr->id != NULL)
    {
      g_free(rgr->id);
      rgr->id = NULL;
    }

  if(rgr->to != NULL)
    {
      g_free(rgr->to);
      rgr->to = NULL;
    }

  if(rgr->from != NULL)
    {
      g_free(rgr->from);
      rgr->from = NULL;
    }

  if(rgr->instructions != NULL)
    {
      g_free(rgr->instructions);
      rgr->instructions = NULL;
    }

  if(rgr->username != NULL)
    {
      g_free(rgr->username);
      rgr->username = NULL;
    }

  if(rgr->password != NULL)
    {
      g_free(rgr->password);
      rgr->password = NULL;
    }

  if(rgr->email != NULL)
    {
      g_free(rgr->email);
      rgr->email = NULL;
    }

  rgr->require_username = FALSE;
  rgr->require_password = FALSE;
  rgr->require_email = FALSE;
  rgr->require_nickname = FALSE;
  
  if(rgr->key != NULL)
    {
      g_free(rgr->key);
      rgr->key = NULL;
    }

  rgr->cb = NULL;
}

/*
 * find 
 */
GjRegistration gj_registration_find_by_id(const gchar *id)
{
  gint index = 0;

  if(id == NULL)
    {
      g_warning("%s: id was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(registration_list);index++)
    {
      GjRegistration rgr  = g_list_nth_data(registration_list, index);

      if(rgr == NULL) continue;
      if(g_strcasecmp(rgr->id, id) == 0) return rgr;
    }

  return NULL;
}

GjRegistration gj_registration_find_by_to(const gchar *to)
{
  gint index = 0;

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(registration_list);index++)
    {
      GjRegistration rgr  = g_list_nth_data(registration_list, index);

      if(rgr == NULL) continue;
      if(g_strcasecmp(rgr->to, to) == 0) return rgr;
    }

  return NULL;
}

GjRegistration gj_registration_find_by_from(const gchar *from)
{
  gint index = 0;

  if(from == NULL)
    {
      g_warning("%s: from was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(registration_list);index++)
    {
      GjRegistration rgr  = g_list_nth_data(registration_list, index);

      if(rgr == NULL) continue;
      if(g_strcasecmp(rgr->from, from) == 0) return rgr;
    }

  return NULL;
}

/*
 * member functions  
 */

/* sets */
gboolean gj_registration_set_id(GjRegistration rgr, const gchar *id)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  if(id == NULL)
    {
      g_warning("%s: id was NULL", __FUNCTION__);
      return FALSE;
    }

  if(rgr->id != NULL)
    {
      g_free(rgr->id);
      rgr->id = NULL;
    }
  
  rgr->id = g_strdup(id);
  return TRUE;
}

gboolean gj_registration_set_to(GjRegistration rgr, const gchar *to)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(rgr->to != NULL)
    {
      g_free(rgr->to);
      rgr->to = NULL;
    }
  
  rgr->to = g_strdup(to);
  return TRUE;
}

gboolean gj_registration_set_from(GjRegistration rgr, const gchar *from)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  if(from == NULL)
    {
      g_warning("%s: from was NULL", __FUNCTION__);
      return FALSE;
    }

  if(rgr->from != NULL)
    {
      g_free(rgr->from);
      rgr->from = NULL;
    }
  
  rgr->from = g_strdup(from);
  return TRUE;
}

gboolean gj_registration_set_instructions(GjRegistration rgr, const gchar *instructions)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  if(instructions == NULL)
    {
      g_warning("%s: instructions was NULL", __FUNCTION__);
      return FALSE;
    }

  if(rgr->instructions != NULL)
    {
      g_free(rgr->instructions);
      rgr->instructions = NULL;
    }
  
  rgr->instructions = g_strdup(instructions);
  return TRUE;
}

gboolean gj_registration_set_username(GjRegistration rgr, const gchar *username)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  if(username == NULL)
    {
      g_warning("%s: username was NULL", __FUNCTION__);
      return FALSE;
    }

  if(rgr->username != NULL)
    {
      g_free(rgr->username);
      rgr->username = NULL;
    }
  
  rgr->username = g_strdup(username);
  return TRUE;
}

gboolean gj_registration_set_password(GjRegistration rgr, const gchar *password)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  if(password == NULL)
    {
      g_warning("%s: password was NULL", __FUNCTION__);
      return FALSE;
    }

  if(rgr->password != NULL)
    {
      g_free(rgr->password);
      rgr->password = NULL;
    }
  
  rgr->password = g_strdup(password);
  return TRUE;
}

gboolean gj_registration_set_email(GjRegistration rgr, const gchar *email)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  if(email == NULL)
    {
      g_warning("%s: email was NULL", __FUNCTION__);
      return FALSE;
    }

  if(rgr->email != NULL)
    {
      g_free(rgr->email);
      rgr->email = NULL;
    }
  
  rgr->email = g_strdup(email);
  return TRUE;
}

gboolean gj_registration_set_require_username(GjRegistration rgr, gboolean required)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }
  
  rgr->require_username = required;
  return TRUE;
}

gboolean gj_registration_set_require_password(GjRegistration rgr, gboolean required)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }
  
  rgr->require_password = required;
  return TRUE;
}

gboolean gj_registration_set_require_email(GjRegistration rgr, gboolean required)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }
  
  rgr->require_email = required;
  return TRUE;
}

gboolean gj_registration_set_require_nickname(GjRegistration rgr, gboolean required)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }
  
  rgr->require_nickname = required;
  return TRUE;
}

gboolean gj_registration_set_key(GjRegistration rgr, const gchar *key)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  if(key == NULL)
    {
      g_warning("%s: key was NULL", __FUNCTION__);
      return FALSE;
    }

  if(rgr->key != NULL)
    {
      g_free(rgr->key);
      rgr->key = NULL;
    }
  
  rgr->key = g_strdup(key);
  return TRUE;
}

gboolean gj_registration_set_callback(GjRegistration rgr, GjRegistrationCallback cb)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  rgr->cb = cb;
  return TRUE;
}

/* gets */
const gchar *gj_registration_get_id(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return NULL;
    }

  return rgr->id;
}

const gchar *gj_registration_get_to(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return NULL;
    }

  return rgr->to;
}

const gchar *gj_registration_get_from(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return NULL;
    }

  return rgr->from;
}

const gchar *gj_registration_get_instructions(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return NULL;
    }

  return rgr->instructions;
}

/* const gchar *registrationGetNickname(GjRegistration rgr) */
/* { */
/*   if(rgr == NULL) */
/*     { */
/*       g_warning("registrationGetNickname: rgr was NULL"); */
/*       return NULL; */
/*     } */

/*   return rgr->nickname; */
/* } */

const gchar *gj_registration_get_password(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return NULL;
    }

  return rgr->password;
}

const gchar *gj_registration_get_email(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return NULL;
    }

  return rgr->email;
}

gboolean gj_registration_get_require_username(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  return rgr->require_username;
}

gboolean gj_registration_get_require_password(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  return rgr->require_password;
}

gboolean gj_registration_get_require_email(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  return rgr->require_email;
}

gboolean gj_registration_get_require_nickname(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return FALSE;
    }

  return rgr->require_nickname;
}

const gchar *gj_registration_get_key(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return NULL;
    }

  return rgr->key;
}

GjRegistrationCallback gj_registration_get_callback(GjRegistration rgr)
{
  if(rgr == NULL)
    {
      g_warning("%s: registration was NULL", __FUNCTION__);
      return NULL;
    }

  return rgr->cb;
}

/*
 *
 * SERVICES
 *
 */
gboolean gj_registration_request_requirements(const gchar *server, 
					      GjRegistrationRequestCallback cb)
{
  if(server == NULL)
    {
      g_warning("%s: server was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  registration_cb = cb;

  /* do request */
  gj_iq_request_registration_requirements(server, gj_registration_cb, NULL);

  return TRUE;
}

gboolean gj_registration_request(const gchar *server, 
				 GjRegistrationRequestCallback cb, 
				 const gchar *key, 
				 const gchar *username, 
				 const gchar *password, 
				 const gchar *email, 
				 const gchar *nickname)
{
  if(server == NULL)
    {
      g_warning("%s: server was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  registration_cb = cb;

  /* do request */
  gj_iq_request_registration(server, gj_registration_cb, 
			     "key", key?key:"",
			     "username", username?username:"",
			     "password", password?password:"",
			     "email", email?email:"",
			     "nickname", nickname?nickname:"",
			     NULL);

  return TRUE;
}

static void gj_registration_cb(guint event_id, guint related_event_id)
{
  GjInfoquery iq = NULL;
  GjRegistration rgr = NULL;

  gint error_code = 0;
  gchar *error_reason = NULL;

  xmlChar *type = NULL;
  xmlNodePtr node = NULL;

  if(registration_cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return;
    }

  if((iq = gj_iq_find(related_event_id)) == NULL)
    {
      g_warning("%s: could not get infoquery from event id:%d", __FUNCTION__, event_id);
      return;
    }

  node = gj_iq_get_xml(iq);
  type = gj_parser_find_by_attr_and_ret(node, (guchar*)"type");

  if(type != NULL && xmlStrcmp(type, (guchar*)"error") == 0)
    {
      xmlChar *errorCode = NULL;
      xmlChar *errorMessage = NULL;

      if((node = gj_parser_find_by_node(node, (guchar*)"error")) == NULL)
	{
	  g_warning("%s: failed to find node 'error'", __FUNCTION__);
	  gj_iq_free_by_event_id(event_id);
	  return;
	}

      errorCode = gj_parser_find_by_attr_and_ret(node, (guchar*)"code");
      errorMessage = gj_parser_find_by_node_and_ret(node, (guchar*)"error");

      error_code = atoi((gchar*)errorCode);
      error_reason = g_strdup((gchar*)errorMessage);
    }

  /* parse list */
  if((rgr = gj_registration_parse(node)) == NULL)
    {
      g_warning("%s: could not parse xml node", __FUNCTION__);
      return;
    }

  /* return list */
  (registration_cb)(rgr, error_code, error_reason);

  /* clean up */
  gj_iq_free_by_event_id(event_id);
  if(error_reason != NULL) g_free(error_reason);
}

static GjRegistration gj_registration_parse(xmlNodePtr parent)
{
  xmlChar *tmp = NULL;
  GjRegistration rgr = NULL;

  if(parent == NULL)
    {
      g_warning("%s: xml node was NULL", __FUNCTION__);
      return NULL;
    }

  /* create new roster item */
  if((rgr = gj_registration_new()) == NULL)
    {
      g_warning("%s: could not create registration", __FUNCTION__);
      return NULL;
    }

  /* get important information */
  if((tmp = gj_parser_find_by_attr_and_ret(parent, (guchar*)"id")) != NULL)
    gj_registration_set_id(rgr, (gchar*)tmp);
  if((tmp = gj_parser_find_by_attr_and_ret(parent, (guchar*)"to")) != NULL)
    gj_registration_set_to(rgr, (gchar*)tmp);
  if((tmp = gj_parser_find_by_attr_and_ret(parent, (guchar*)"from")) != NULL)
    gj_registration_set_from(rgr, (gchar*)tmp);
  
  /* get instructions */
  if((tmp = gj_parser_find_by_node_and_ret(parent, (guchar*)"instructions")) != NULL)
    gj_registration_set_instructions(rgr, (gchar*)tmp);

#if 0
  /* get user/pass... */
  if((tmp = gj_parser_find_by_node_and_ret(parent, (guchar*)"username")) != NULL)
    gj_registration_set_username(rgr, (gchar*)tmp);
  if((tmp = gj_parser_find_by_node_and_ret(parent, (guchar*)"password")) != NULL)
    gj_registration_set_password(rgr, (gchar*)tmp);
  if((tmp = gj_parser_find_by_node_and_ret(parent, (guchar*)"email")) != NULL)
    gj_registration_set_email(rgr, (gchar*)tmp);
#endif

  /* get user/pass... */
  if(gj_parser_find_by_node(parent, (guchar*)"username") != NULL)
    gj_registration_set_require_username(rgr, TRUE);
  if(gj_parser_find_by_node(parent, (guchar*)"password") != NULL)
    gj_registration_set_require_password(rgr, TRUE);
  if(gj_parser_find_by_node(parent, (guchar*)"email") != NULL)
    gj_registration_set_require_email(rgr, TRUE);
  if(gj_parser_find_by_node(parent, (guchar*)"nickname") != NULL)
    gj_registration_set_require_nickname(rgr, TRUE);

  if((tmp = gj_parser_find_by_node_and_ret(parent, (guchar*)"key")) != NULL)
    gj_registration_set_key(rgr, (gchar*)tmp);

  return rgr;
}

#ifdef __cplusplus
}
#endif
